#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 18 18:44:09 2020

@author: mario
"""
import networkx as nx
import gurobipy as gp
from gurobipy import GRB
import datetime
import sys
import matplotlib.pyplot as plt
import lp_to_nx_graph
from itertools import combinations

    
class DominatingSet:
    
    def __init__(self, G, name = "DS"):
        self.G = G
        self.m = gp.Model(name)
        self.nodes = self.m.addVars(G.nodes, vtype = GRB.BINARY, name = "nodes")
        self.m.setObjective(gp.quicksum(self.nodes), GRB.MINIMIZE)
    
        # For each node in the graph the node itself or at least one of it's neighbors has to be part of the dominating set
        self.m.addConstrs(((gp.quicksum(self.nodes[n] for n in G.neighbors(v)) + self.nodes[v] )>= 1) for v in G.nodes)

    def solve(self):
        ds = {i for i,x_i in enumerate(self.m.getVars()) if x_i.x == 1}
        return ds
    
    def solve_and_draw(self):
        starttime = datetime.datetime.now()
        ds = self.solve()
        endtime = datetime.datetime.now()
        duration = endtime- starttime
        duration_sec = duration.total_seconds()
        color_map = ['red' if i in ds else 'green' for i in self.G.nodes]
        print(f"duration in seconds: {duration_sec}")
        nx.draw_kamada_kawai(self.G, node_color = color_map, with_labels = True)
        plt.show()
        
    
class KHopDominatingSet(DominatingSet):
    
    def __init__(self, G, k, name = "kDS"):
        super().__init__(KHopDominatingSet.make_closure(G,k), name)
        self.G = G

    def make_closure(G, k):
        G_prime = nx.Graph(G)
        for v in G.nodes:
            neighbors = set(G.neighbors(v))
            for i in range(k-1):
                neighbors.update([w for n in neighbors for w in G.neighbors(n)])
            
            G_prime.add_edges_from((v,n) for n in neighbors)
        
        return G_prime

class ConnectedKHopDominatingSet(KHopDominatingSet):
    
    def __init__(self, G, k, name = "CkDS", exclude = {}):
        self.G = G
        super().__init__(G, k, name)
        
        if exclude:
            self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes if v not in exclude)
        else:
            self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes)
        

        
    def min_ij_separator(G, i, j, C_i):
        N_ci = {v for c in C_i for v in G.neighbors(c)}
        G_prime = nx.Graph(G)
        C_i_prime = C_i.copy()
        C_i_prime.update(N_ci)
        G_prime.remove_edges_from(G.subgraph(C_i_prime).edges)
        # dijkstra
        R_j = nx.algorithms.dag.descendants(G_prime, j)
        return R_j.intersection(N_ci)
    
    def solve(self):
        self.m._vars = self.nodes
        self.m._G = self.G
        self.m.Params.lazyConstraints = 1
        self.m.optimize(RootedConnectecKHopDominatingSet.elim_unconnectivity)    
        # self.m.optimize()
        
        ds = {i for i,x_i in enumerate(self.m.getVars()) if x_i.x > 0.5}
        return ds
    
    def elim_unconnectivity(model, where):
        if where == GRB.Callback.MIPSOL:
            vals = model.cbGetSolution(model._vars)
            ds = {i for i in model._vars.keys() if vals[i] > 0.5}
            
            G_prime_prime = model._G.subgraph(ds)
            if(not nx.is_connected(G_prime_prime)):
                C = [c for c in nx.algorithms.components.connected_components(G_prime_prime)]
                for i in range(len(C)-1):
                    C_i = C[i]
                    for j in range(i+1, len(C)):
                        C_j = C[j]
                        h = next(iter(C_i))
                        l = next(iter(C_j))
                        min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(model._G, h, l, C_i)
                        model.cbLazy(gp.quicksum(model._vars[s] for s in min_ij_sep) >= model._vars[h] + model._vars[l] - 1)

        
class RootedConnectecKHopDominatingSet(ConnectedKHopDominatingSet):
    
    def __init__(self, G, k, root = 0, name = "RCkDS"):
        super().__init__(G, k, name, exclude = {root})
        self.root = root
        
        self.m.addConstr(self.nodes[root] >= 1)
        
    def add_single_root_separators(self):
        for i in self.G.nodes:
            min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(G, i, self.root, {i})
            if min_ij_sep:
                self.m.addConstr(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[i])
        
    # ToDo: refactor!
    def add_all_neighborhood_root_separators(self):
        for v in self.G.nodes:
            if v is not self.root and self.root not in self.G.neighbors(v) and v not in self.G.neighbors(self.root) and not set(self.G.neighbors(self.root)).intersection(set(self.G.neighbors(v))):
                for i in range(2,self.G.degree[v]):
                    V = {w for w in self.G.neighbors(v)}
                    V.update([v])
                    for i_neighborhood in combinations(V, i):
                        if v in i_neighborhood:
                            min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(self.G, v, self.root, set(i_neighborhood))
                            self.m.addConstr(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[v])
    
    def add_all_combinations_root_separators(self):
        for v in self.G.nodes:
            for i in range(2,len(self.G.nodes)):
                V = {w for w in self.G.neighbors(v)}
                V.update([v])
                for i_neighborhood in combinations(V, i):
                    if v in i_neighborhood:
                        min_ij_sep = ConnectedKHopDominatingSet.min_ij_separator(self.G, v, self.root, set(i_neighborhood))
                        if min_ij_sep:
                            self.m.addConstr(gp.quicksum(self.nodes[s] for s in min_ij_sep) >= self.nodes[v])
        
    def add_simple_path_length_constraint(self):
        self.m.addConstrs((self.nodes[v] * len(nx.algorithms.shortest_path(self.G, self.root, v))) <= gp.quicksum(self.nodes) for v in self.G.nodes)
        # self.m.addConstrs((self.nodes[v] * self.nodes[w] * len(nx.algorithms.shortest_path(self.G, v, w))) <= gp.quicksum(self.nodes) for v in self.G.nodes for w in self.G.nodes)

        
    def add_gausian_sum_formula_constraint(self):
        self.m.addConstr(gp.quicksum(self.nodes[v] * len(nx.algorithms.shortest_path(self.G, self.root, v)) for v in self.G.nodes) <= (gp.quicksum(self.nodes)+1)*gp.quicksum(self.nodes)/2)
        # self.m.addConstr(gp.quicksum(self.nodes[v] * self.nodes[w] * len(nx.algorithms.shortest_path(self.G, v, w)) for v in self.G.nodes for w in self.G.nodes) <= (gp.quicksum(self.nodes)+1)*gp.quicksum(self.nodes)/2)

    def add_neighbor_of_neighbors_constraint(self, exclude):
        self.m.addConstrs(self.nodes[v] <= gp.quicksum(self.nodes[w] * gp.quicksum(self.nodes[h] for h in self.G.neighbors(w) if h is not v) for w in self.G.neighbors(v)) for v in self.G.nodes if v not in exclude)

    def intermediate_node_constraint(self, exclude):
        self.m.addConstrs(2 * self.nodes[v] <= gp.quicksum(self.nodes[w] for w in G.neighbors(v)) for v in G.nodes if v not in exclude)


if __name__ == '__main__':
    G = lp_to_nx_graph.read(sys.argv[1])
    
    if(len(sys.argv) > 2):
        k = int(sys.argv[2])
    else:
        k = 1
        
    
    # G = lp_to_nx_graph.read("/home/mario/Dokumente/Uni/Bachelorarbeit/git_repo/bachelor-mario-surlemont/python/lp graphs/asymmetric.lp")
    # k = 2
    dsProb = RootedConnectecKHopDominatingSet(G, k, 0)
    dsProb.solve_and_draw()
